/*
 * Author: Michael Jonas (copyright 2020) released under GNU GPL v3
 * Date: 3/18/20
 * ZXSimulator: a prototype ZX80/81 BASIC Simulator
 */


#include stdio_h


/****************************** cmd_h *******************************/

#define ZXBREAK    keyrow(7)==2) if (keyrow(3)==8          /* ctl-s */
#define getexpr    evalbool    /* top expression in precedence tree */

#define addr       long

#define EOL       '\n'
#define EOP       '\0'
#define LINELEN    128  /* was 256 */
#define LINEMAX    125  /* 3 less from LINELEN for EOL/EOF or cursor */
#define CMDLEN     20
#define VARLEN     20
#define BLKLEN     64
#define NUMLEN     10
#define MAXDIM     6
#define MAXLNUM    9999
#define MAXINT     32768

#define BLKNEST    10
#define VARSIZE    16
#define CREATE     0
#define RESET      1

#define NOERR      0
#define NOCMD     -1
#define NOVAL     -2
#define NOHEAP    -3
#define MAXLIM    -4
#define BADARG    -5
#define BADCHAR   -6
#define NOMORE    -7
#define UNKNOWN   -8

/* constant for PI */
#define PI         3  /* (3.1415927) */

/* constants for random number generator */
#define P1         8421
#define P2         1
#define N          65537

/* constants for ZX graphics */
#define WINTOP     1
#define WINSIDE    124    /* (512-(MAXCOL+4))/2 */
#define WINWIDTH   264    /* MAXCOL+4 */
#define WINHEIGHT  192    /* MAXROW+4 */
#define FIRSTROW   2
#define MAXROW     188
#define LASTROW    170
#define CURSROW    180
#define FIRSTCOL   2
#define MIDCOL     130
#define LASTCOL    250
#define MAXCOL     260
#define COLLEN     32
#define MAXPLOTY   43
#define MAXPLOTX   63

/* character map */
char sym[17] = {'"','#','$',':','?','(',')','>','<','=','+','-','*','/',';',',','.'};

/* global command variables */
int  globerr = NOERR;
char ws[14] = {' ',',',';','\n','(',')','>','<','=','+','-','*','/','\0'};
char spc[5] = {' ',',',';','\n','\0'};
char *delim;
int  doSTR;
char cstr[LINELEN] = "";
char *pstr;
int  mlen = 0;
int  plen;
char *str;
int  sp;

/* data sturctures for variables */
addr *var_nam;
addr *var_ptr;
int  *var_num;
int  var_len;

/* data structures for loops gosubs, and if statements */
int  *lp_ptr;
int  *lp_end;
int  *lp_inc;
addr *lp_var;
int  lp_len;
int  lp_cnt = 0;

int  *rt_ptr;
int  rt_len;
int  rt_ind = -1;

long key[8] = {0X0};
long ret[8];

/* random seed */
long seed;

/* data structurs for ZX graphics */
int  col = FIRSTCOL;
int  row = FIRSTROW;
int  endrow = LASTROW;
int  dquote;
int  inprow;


/****************************** cmd_c *******************************/

main ()
{
   char cmd[CMDLEN];
   int err;

   initPARSE();

   while (1)
   {
      if (globerr != NOERR)
         { globerr = NOERR; clearpgm(); }

      reset();

      /* get next input from user */
      ZXgets(cstr, 0);

      /* list program on empty command */
      if (cstr[0] == EOL)
         { listpgm(0); continue; }

      /* add BASIC line: NOERR if ok, NOVAL if not, error otherwise */
      err = addline();
      if (err == NOERR) continue;

      /* otherwise we need to parse the command */
      if (err == NOVAL)
         err = PARSEcmds(cmd);
     
      /* and report any errors */
      switch (err)
      {
         case NOERR  :
         case NOMORE : break;
         case MAXLIM : showerr("MAX LIMIT"); break;
         case NOCMD  : showerr("NO COMMAND"); break;
         case BADARG : showerr("BAD ARGUMENT"); break;
         default     : showerr("UNKNOWN");
      }
   }
}


/* initializes the parser */
initPARSE ()
{
   int i;

   initQLscr();
   seed = date();
   key[4] = get_chanid(stdin);

   var_len=VARSIZE;
   initvar(CREATE);

   rt_len=BLKNEST;
   memarg(&rt_ptr, BLKNEST*2);

   lp_len=BLKNEST;
   initlp(CREATE);

   clearpgm();
}


/* parser */
PARSEcmds(cmd)
   char cmd[];
{
   int  err, ln;

   while (1)
   {
      if (ZXBREAK) return NOERR;

      if (getcmd(cmd) != NOERR) return BADARG;
      if ((err=BASICcmds(cmd)) != NOERR) return err;

      if (str[sp] == EOL)        /* go to next line */
         { sp++; linenum(&ln); }
      if (str[sp] == EOP) return NOERR;
   }
}


/* BASIC commands */
BASICcmds (cmd)
   char cmd[];
{
   char ch, *val_s, *vl2_s, *ostr, arg[LINELEN], var[VARLEN];
   int  val_i, vl2_i, vl3_i, vl4_i, err, op, i, dim[MAXDIM];
   addr mem;

   if (streql("CLEAR",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      initvar(RESET); initlp(RESET);
   }
   else if (streql("CLS",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      cls(stdout);
      row = FIRSTROW; col = FIRSTCOL;
   }
   else if (streql("DIM",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if ((ch=var[strlen(var)-1]) == '(') sp--;
      if (getsym("(") != NOERR) return BADARG;
      vl2_i = 0; vl3_i = 1;
      do { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (vl2_i < MAXDIM) dim[vl2_i++] = val_i; else return MAXLIM;
           if (val_i >= 1) vl3_i *= val_i; else return BADARG;
      } while (getsym(",") == NOERR);
      if (getsym(")") != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;

      if (ch == '$')
         { if (vl2_i > 1) return MAXLIM;
           memarg(&val_s, val_i+1);
           i=0; while (i<val_i) { setchp(val_s,i++,' '); } setchp(val_s,i,'\0');
           err = storevar(var,0,0,val_s); free(val_s);
           return err; }
      else
         { if (findvar(var,&i) == NOVAL)
              { memarg(&mem,strlen(var)+1); var_nam[i]=mem; strcpy(mem,var); }
           else
              { free(var_ptr[i]); }

           memarg(&mem,(vl2_i+vl3_i+1)*2); var_ptr[i] = mem; 
           setintp(mem,0,vl2_i); i = 1;
           while (i<=vl2_i) { setintp(mem,i,dim[i-1]); i++; }
           while (i<=vl2_i+vl3_i) setintp(mem,i++,0); }
   }
   else if (streql("EDIT",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      return editline(val_i);
   }
   else if (streql("FOR",cmd))
   {
      if (getvar(var) != NOERR || getsym("=") != NOERR) return BADARG;
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("TO") != NOERR) return BADARG;
      if (getexpr(&vl2_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("STEP") != NOERR)
         vl3_i = 1;
      else if (getexpr(&vl3_i,&val_s) == NOERR)
         free(val_s);
      else
         return BADARG;
      if (noarg() != NOERR) return BADARG;
      if ((err=nestlp(var,val_i,vl2_i,vl3_i)) == NOVAL)
         { while (str[sp] != EOF)
              if (str[sp++] == EOL)
                 { while ((ch=str[sp]) != ' ') { if (ch == EOF) break; else sp++; }
                   if (getkey("NEXT") == NOERR) if (getvar(arg) == NOERR) if (streql(var, arg))
                      break; } }
      else
         { return err; }
   }
   else if (streql("GOSUB",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      if ((++rt_ind) >= rt_len) 
         { memarg(&mem,(rt_len+BLKNEST)*2); i=0; while (i<rt_len) { setintp(mem,i,rt_ptr[i]); i++; }
           free(rt_ptr); rt_ptr=mem; rt_len += BLKNEST; }
      rt_ptr[rt_ind] = sp; findline(val_i);
   }
   else if (streql("GOTO",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      findline(val_i);
   }
   else if (streql("IF",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getkey("THEN") != NOERR) return BADARG;
      if (noarg() == NOERR) return BADARG;
      if (!val_i) skipln();
   }
   else if (streql("INPUT",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (var[strlen(var)-1] == '(')
         { if (getindex (var,&val_i,&val_s) == NOERR) free(val_s); else return BADARG; }
      else
         { val_i = 0; }
      if (noarg() != NOERR) return BADARG;

      ZXgets(arg,1);
      return storevar(var,val_i,atoi(arg),arg);
   }
   else if (streql("LET",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (var[strlen(var)-1] == '(')
         { if (getindex (var,&vl4_i,&val_s) == NOERR) free(val_s); else return BADARG; }
      else
         { vl4_i = 0; }

      if (getsym("=") != NOERR)
         { if (fetchvar(var,0,&val_i,&val_s) != NOERR) return BADARG;
           if (getrange(&vl2_i,&vl3_i,strlen(val_s)) != NOERR) goto free_on_err;
           if (getsym("=") != NOERR) goto free_on_err;
           if (getexpr(&val_i,&vl2_s) != NOERR) goto free_on_err;
           i=vl2_i; while (i<=vl3_i) { if ((ch=vl2_s[i-vl2_i])=='\0') break; val_s[i-1]=ch; i++; }
           val_i=atoi(val_s); free(vl2_s); }
      else
         { if (getexpr(&val_i,&val_s) != NOERR) return BADARG; }

      if (noarg() != NOERR) { free_on_err: free(val_s); return BADARG; }

      err = storevar(var,vl4_i,val_i,val_s); free(val_s);
      return err;
   }
   else if (streql("LIST",cmd))
   {
      if (noarg() != NOERR)
         { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           op = sp; ostr = str; findline(val_i); }
      else
         { op = sp; ostr = str; sp = 1; }

      if (listpgm(0) != NOERR) return NOMORE;
      sp = op; str = ostr;
   }
   else if (streql("LOAD",cmd))
   {
      if (getexpr(&val_i,&val_s) != NOERR) return BADARG;
      if ((err=noarg()) == NOERR) err = loadfile(val_s);
      free (val_s); if (err != NOERR) return BADARG;
      initvar(RESET); reset();
   }
   else if (streql("NEW",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      clearpgm(); initvar(RESET); reset();
   }
   else if (streql("NEXT",cmd))
   {
      if (getvar(var) != NOERR) return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (nextlp(var) != NOERR) return BADARG;
   }
   else if (streql("PAUSE",cmd))
   {
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      while (1) { pause(5); if ((val_i-=5)<=0) break; trap3(key, ret); if (!ret[0]) break; }
   }
   else if (streql("PLOT",cmd))
   {
      i = 0;
      plot_command:
      if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (getsym(",") != NOERR) return BADARG;
      if (getexpr(&vl2_i,&val_s) == NOERR) free(val_s); else return BADARG;
      if (noarg() != NOERR) return BADARG;
      if (val_i > MAXPLOTX || vl2_i > MAXPLOTY) return BADARG; 
      block(stdout, 4, 4, FIRSTCOL+(abs(val_i)<<2), LASTROW+4-(abs(vl2_i)<<2), i);
   }
   else if (streql("UNPLOT",cmd))
   {
      i = 7;
      goto plot_command;
   }
   else if (streql("PRINT",cmd))
   {
      return putformat();
   }
   else if (streql("RAND",cmd))
   {
      if (noarg() != NOERR)
         { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           if (val_i) { seed = val_i; return NOERR; } }
      seed = date();
   }
   else if (streql("REM",cmd))
   {
      skipln();
   }
   else if (streql("RETURN",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      if (rt_ind < 0) return BADARG;
      sp = rt_ptr[rt_ind--];
   }
   else if (streql("RUN",cmd))
   {
      if (noarg() != NOERR)
         { if (getexpr(&val_i,&val_s) == NOERR) free(val_s); else return BADARG;
           if (noarg() != NOERR) return BADARG;
           reset(); findline(val_i); }
      else
         { reset(); str = pstr; }
      initvar(RESET);
   }
   else if (streql("SAVE",cmd))
   {
      if (getexpr(&val_i,&val_s) != NOERR) return BADARG;
      if ((err=noarg()) == NOERR) err = savefile(val_s);
      free (val_s); if (err != NOERR) return BADARG;
   }
   else if (streql("SCROLL",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      scroll(stdout, -8);
      block(stdout, MAXROW, 1, 2, 0, 7); /* clear the top line */
      row = LASTROW; col = FIRSTCOL;
   }
   else if (streql("STOP",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      reset();
   }
   else if (streql("QUIT",cmd))
   {
      if (noarg() != NOERR) return BADARG;
      abort(0);
   }
   else if (cmd[0] != '\0')        /* if command is not empty */
   {
      if (!streql("FAST",cmd))     /* list of commands we ignore */
      if (!streql("SLOW",cmd))
         return NOCMD;
   }

   return NOERR;
}


/****************************** aux_c *******************************/

/* gets command from command line */
getcmd (cmd)
   char cmd[];
{
   char ch;
   int  i;

   while (str[sp] == ' ') sp++;
   i=0;
   while (i<CMDLEN)
      { if ((ch=str[sp]) != ' ') if (ch != EOL)
           { cmd[i++] = ch; sp++; continue; }
        cmd[i] = '\0'; return NOERR; }

   return BADARG;
}   /* end getcmd */


/* gets keyword from command line */
getkey (key)
   char key[];
{
   char ch;
   int i, op;

   while (str[sp] == ' ') { sp++; } op = sp;
   i=0;
   while ((ch=key[i++])!='\0')
      if (str[sp++] != ch) goto bad_key;
   if (strchr(delim,str[op-1]) != NULL) if (strchr(delim,str[sp]) != NULL)
      return NOERR;

   bad_key: sp = op;

   return BADCHAR;
}


/* gets symbol from command line */
getsym (sym)
   char sym[];
{
   char ch;
   int i, op;

   while (str[sp] == ' ') { sp++; } op = sp;
   i=0;
   while ((ch=sym[i++])!='\0')
      if (str[sp++] != ch) { sp=op; return BADCHAR; }

   return NOERR;
}


/* displays formatted argument for print statement */
putformat ()
{
   int val_i, err, doLF;
   char *val_s;

   err = NOERR;
   doLF = -1;   /* neither no-line-feed or yes-line-feed so print is ok */
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case 'A':
            if (getkey("AT") != NOERR) break;
            if (doLF > 0 || getexpr(&val_i,&val_s) != NOERR) { return BADARG; } free(val_s);
            row = FIRSTROW + (abs(val_i) << 3);
            if (getsym(",") != NOERR) return BADARG;
            if (getexpr(&val_i,&val_s) != NOERR) { return BADARG; } free(val_s);
            col = FIRSTCOL + (abs(val_i) << 3);
            if (row > LASTROW || col > LASTCOL) return BADARG;
            doLF = 2;
            continue;
         case 'T':
            if (getkey("TAB") != NOERR) break;
            if (doLF > 0 || getexpr(&val_i,&val_s) != NOERR) { return BADARG; } free(val_s);
            if (val_i < 0) return BADARG;
            if ((col >> 3) > val_i) row = row + 8;
            col = FIRSTCOL + ((val_i % COLLEN) << 3);
            doLF = 2;
            continue;
         case ',':
            if (col < MIDCOL)
               { col = MIDCOL; if (row > endrow) return NOMORE; }
            else
               { ZXdowncurs(); }
         case ';':
            sp++; doLF = 0;
            continue;
         case EOL:
            switch (doLF)
            {
               case 0:
                  break;
               case 1:
                  if (col == FIRSTCOL) break;
               default:
                  ZXdowncurs();
            }
            return err;
      }
      if (doLF > 0 || getexpr(&val_i,&val_s) != NOERR) return BADARG;
      err = ZXprintf(val_s);
      free(val_s); doLF = 1;
   }
}


/* checks for no argument from command line */
noarg ()
{
   while (str[sp] == ' ') sp++;
   if (str[sp] != EOL)
      return BADCHAR;
   else
      return NOERR;
}


/* skips to end of line */
skipln ()
{
   while (str[sp] != EOL) sp++;
}


/* adds new loop nest for "var" */
nestlp (name,start,stop,step)
   char name[];
   int start, stop, step;
{
   char ch;
   addr mem, *var;
   int  i, *ptr, *end, *inc;

   if ((ch=name[strlen(name)-1])=='$' || ch=='(') return BADARG;

   i = 0;
   while (i < lp_cnt)
      if (streql(lp_var[i],name)) break; else i++;

   /* if limit reached, resize loop structure */
   if (i == lp_len)
   {
      ptr=lp_ptr; end=lp_end; inc=lp_inc; var=lp_var;
      i=0; lp_len+=BLKNEST; initlp(CREATE);
      while (i<lp_cnt)
         { lp_ptr[i]=ptr[i]; lp_end[i]=end[i]; lp_inc[i]=inc[i]; lp_var[i]=var[i]; i++; }
      free(ptr); free(end); free(inc); free(var);
   }

   storevar(name,0,start,"");

   if (step > 0) if (start > stop) return NOVAL;
   if (step < 0) if (start < stop) return NOVAL;

   lp_ptr[i] = sp;
   lp_end[i] = stop;
   lp_inc[i] = step;
   if (i == lp_cnt)
   {
      memarg(&mem,strlen(name)+1);
      lp_var[i] = mem; strcpy(mem,name);
      lp_cnt++; 
   }
      
   return NOERR;
}


/* skips to next increment of "var" */
nextlp (name)
   char name[];
{
   char ch;
   addr val_s;
   int i, val_i, step, stop;

   if ((ch=name[strlen(name)-1])=='$' || ch=='(') return BADARG;

   i = lp_cnt;
   while (i > 0)
      if (streql(lp_var[--i],name))
      {
         fetchvar(name,0,&val_i,&val_s);
         free(val_s); step = lp_inc[i];  stop = lp_end[i];
         val_i += step; storevar(name,0,val_i,"");

         if (step > 0) if (val_i > stop) return NOERR;
         if (step < 0) if (val_i < stop) return NOERR;
         sp=lp_ptr[i]; return NOERR;
      }

   return BADARG;
}


/* list program */
listpgm (ed)
   int ed;
{
   int i, j, cp;
   char ch, lstr[LINELEN];

   dquote = 1; /* show double quote: "" */

   i=0; cp=sp;
   while (i<=(ed*10))   /* if editing, new line to middle of screen */
      { i++; if (cp > 1) { do cp--; while (pstr[cp-1]!=EOL); } else { cp=1; break; } }

   j = -1;
   while (pstr[cp]!=EOP)
   {
      if (j == -1)
      {
         j = 0;
         if (pstr[cp+1] == ' ')
            { lstr[j++] = ' '; goto in2; }  /* reduce code to indent 3 */
         else if (pstr[cp+2] == ' ')
            { in2: lstr[j++] = ' '; goto in3; } /* reduce code to indent 2  */
         else if (pstr[cp+3] == ' ')
            { in3: lstr[j++] = ' '; }
         i = j;
      }

      if (ed) if (cp == sp)
         { lstr[j++] = '%'; lstr[j++] = '>'; cp++; }

      lstr[j++] = pstr[cp];

      if ((ch=pstr[cp++]) == '\\')
         { lstr[j++] = pstr[cp++]; goto skipch; }
      if (ch == '%')
         { skipch: lstr[j++] = pstr[cp++]; }
  
      if (i != COLLEN) if (ch != EOL) { i++; continue; }

      if (i == COLLEN) if (ch == EOL) { j--; } lstr[j] = '\0';
      if (ZXprintf(lstr) != NOERR) return NOMORE;
      i = 1; j = 0 - (ch == EOL);

      trap3(key, ret);  /* get next key input */
      if (!ret[0]) if (ZXBREAK) break;
   }

   dquote = 0; /* show single quote: " */
   return NOERR;
}


/* saves program to file */
savefile (fname)
   char fname[];
{
   char ch;
   int fp, i;

   i=0;
   while ((ch=fname[i])!='\0') fname[i++] = tolower(ch); /* lowcase name */
   if ((fp=fopen(fname,"w")) == NULL) return BADARG;

   i=1;
   while ((ch=pstr[i++])!=EOP)
   {
      fputc(ch,fp);

      /* we change our intneral \"" to \" to be compatible */
      if (ch == '"') if (pstr[i] == '"') if (pstr[i-2] == '\\') i++;
   }
       
   fclose(fp);

   return NOERR;
}


/* loads file into string */
loadfile (fname)
   char fname[];
{
   int fp, l, ol, ln, ch, err;

   if ((fp=fopen(fname,"r")) == NULL) return BADARG;
 
   if (setpmem(fp) != NOERR) return NOMORE;

   l = 0; ln = 0; clearpgm();
   while ((err=checkline(fp,&ln,&l)) == NOERR)
      if ((err=readline(fp,&l)) != NOERR) break;

   if (err != BADCHAR)
      { if (pstr[l] != EOL) { setpstr(++l,EOL); } setpstr(++l,EOP); plen=l; }
   else
      { ol=l; readline(fp,&l); pgmerr("BAD LINE",ol); clearpgm(); }
   fclose(fp);
   return NOERR;
}


/* allocates file buffer for program */
setpmem (fp)
   int fp;
{
   long size;

   size = BLKLEN;
   while (fseek(fp, size, 0) == 0) size+=BLKLEN;
   fseek(fp, 0L, 0);

   if (mlen) free(pstr); mlen = size;
   memarg(&pstr, mlen);
   if (pstr == NULL) { mlen = 0; return NOMORE; }

   return NOERR;
}

  
/* reads character into program string, realloc's memory */
setpstr (l,ch)
   int l;
   char ch;
{
   char *ostr, *pmem;
   int  i;

   if (l >= mlen)
   {
      memarg(&pmem, l+BLKLEN);
      if (pmem == NULL) return;

      ostr = pstr; pstr = pmem;
      i=0; while (i<mlen) pstr[i] = ostr[i++];
      if (mlen) free(ostr);
      mlen = l + BLKLEN;
   }
   pstr[l] = ch;
}


/* checks line number from file */
checkline (fp,ln,l)
   int fp, *ln, *l;
{
   char lnum[NUMLEN];
   int  i, num, ch, err;

   i = 0;
   while (((ch=fgetc(fp)) == ' ') || (ch == EOL));
   if (ch == EOF) return EOF;
   while (isdigit(ch))
   {
      lnum[i++] = ch; setpstr(++(*l),ch);
      if ((ch=fgetc(fp)) == EOF) { (*l) -= i; return EOF; }
   }

   lnum[i] = '\0'; num = atoi(lnum); setpstr(++(*l),ch);
   if (ch == ' ') if (num > *ln) if (num <= MAXLNUM)
      { *ln = num; return NOERR; }
   
   return BADCHAR;
}


/* reads line from file */
readline (fp,l)
   int fp, *l;
{
   int ch;

   if (pstr[*l] != EOL)
   {
      while ((ch=toupper(fgetc(fp))) != EOL)
      {
         if (ch == '\\')     /* we change \" to our intneral \"" to be compatible */
            { setpstr(++(*l),ch); if ((ch=toupper(fgetc(fp))) == '"') setpstr(++(*l),ch); }
         if (ch == EOF) return EOF; 
         if (ch != '\r') setpstr(++(*l),ch);  /* get rid of Windows carriage returns */
      }
      setpstr(++(*l),EOL);
   }
   return NOERR;
}


/* finds line number or gives error */
findline (num)
   int num;
{
   int lnum, beg, end, bp, ep;

   str=pstr; beg=1; end=plen;
   while (beg < end)
   {
      sp = (beg + end) >> 1;   /* divide by 2 */

      getline (sp, &bp, &ep); sp=bp;

      if (linenum(&lnum) != NOERR)
         break;

      if (num == lnum)
         return NOERR;

      if (num < lnum)
         end = bp-1;
      else 
      {
         beg = ep+1;
         sp = beg;
      }
   }

   while (isdigit(str[sp])) sp++;

   return NOVAL;
}


/* gets line from program */
getline (lp, bp, ep)
   int lp, *bp, *ep;
{
   char ch;

   *bp=lp; while (*bp) if (pstr[(*bp)-1] != EOL) (*bp)--; else break;
   *ep=lp; while ((ch=pstr[*ep]) != EOP) if (ch != EOL) (*ep)++; else break;
}


/* gets line number from command line */
linenum (ln)
   int * ln;
{
   char ch, num[NUMLEN];
   int  i;

   i=0;
   while(str[sp] == ' ') sp++;
   while(isdigit(ch=str[sp])) { num[i++]=ch; sp++; }
   num[i] = '\0'; *ln = atoi(num);
   if (i) return NOERR; else return NOVAL;
}


/* inserts line into program */
insertline (num)
   int num;
{
   char ch;
   int err, cl, cp, np, bp, ep;

   if ((num < 1) || (num > MAXLNUM)) return BADARG;
   cp=sp; while (cstr[cp]==' ') cp++;

   /* first delete line if found */
   err=findline(num); getline(sp,&bp,&ep);
   if (err == NOERR)
      { np=bp; while (pstr[ep] != EOP) { setpstr(np++, pstr[++ep]); } plen=np-1; }

   ep=plen;

   /* if new line is not empty, add it */
   if (cstr[cp] != EOL)
   {
      cp=0; while (((ch=cstr[cp])==' ') || (ch=='0')) cp++;
      cl=strlen(cstr)-cp; while (ep >= bp) setpstr(ep+cl, pstr[ep--]);
      while ((ch=cstr[cp++])!='\0') setpstr(bp++, ch);
      plen+=cl;
   }

   return NOERR;
}


/* adds new line to program and lists it */
addline ()
{
   int num, err;

   sp = 0; str = cstr;

   if (linenum(&num) != NOERR)
      return NOVAL;

   if (strchr(" \n", str[sp]) != NULL) if (insertline(num) == NOERR)
      { findline(num); listpgm(1); return NOERR; }

   return BADARG;
}


/* edits line of program */
editline (num)
   int num;
{
   int  i, err, bp, ep;

   findline(num);
   if (pstr[sp]==EOP) return NOERR;

   getline(sp,&bp,&ep);

   i = 0;
   while (bp < ep)
      { cstr[i++] = pstr[bp++];
        if (i > LINEMAX) return MAXLIM; }
   cstr[i] = '\0';

   ZXgets(cstr, 0);

   err = addline();
   if (err == NOVAL)
      { listpgm(1); err = NOERR; }
   reset();

   return err;
}


/* points to error in program line */
pgmerr (msg,lp)
   char msg[]; int  lp;
{
   char ch, estr[LINELEN];
   int  i, bp, ep;

   getline(lp,&bp,&ep);

   i = 0;
   while (bp < ep)
      { estr[i++] = pstr[bp++];
       if (lp == bp) { estr[i] = '\0'; lp = ZXlen(estr); }
       if (i > LINEMAX) break; }
   estr[i] = '\0';
   if (lp > i) lp = ZXlen(estr);

   printerr(msg,estr,lp);
}


/* points to error in command line */
showerr (msg)
   char msg[];
{
   dquote = 1;
   if (str == cstr)
      printerr(msg,"",sp);
   else
      pgmerr(msg,sp);
}


/* displays error message */
printerr (msg,estr,pt)
   char msg[], estr[]; int pt;
{
   int len;

   cls(stdout);
   row = FIRSTROW; col = FIRSTCOL;

   if (estr[0] != '\0')
   {
      ZXprintf(estr);
      ZXdowncurs();

      pt = pt % COLLEN;
      col = FIRSTCOL + 8 * pt;

      if ((pt-3) < (len=strlen(msg)))
         { ZXprintf("<--"); }
      else
         { col -= 24; ZXprintf("-->"); col -= 8 * (len + 3); }
   }
   ZXprintf(msg);
}


/* dynamic memory allocation of arg */
memarg (arg,len)
   addr *arg; int len;
{
   if (globerr==NOERR) if (len>0) if ((*arg=malloc(len)) != NULL) return;

   *arg = NULL;
   syserr(NOHEAP);
}


/* sets char ptr to ch */
setchp (chp,ind,ch)
   char chp[]; int ind; char ch;
{
   chp[ind] = ch;
}


/* gets ch from char ptr */
getchp (chp,ind)
   char chp[]; int ind;
{
   return chp[ind];
}


/* sets int ptr to num */
setintp (intp,ind,num)
   int intp[]; int ind; int num;
{
   intp[ind] = num;
}


/* gets num from int ptr */
getintp (intp,ind)
   int intp[]; int ind;
{
   return intp[ind];
}


/* gives system error */
syserr (err)
   int err;
{
   if (globerr == NOERR)
   {
      switch(globerr=err)
      {
         case NOHEAP : showerr("BAD HEAP QUERY"); break;
         default     : showerr("SYSTEM ERROR");
      }
      reset();
   }
}


/* reset input strings */
reset ()
{
   delim = spc; dquote = 0;
   cstr[0] = '\0'; str = cstr; sp = 0;
   rt_ind = -1; initlp(RESET);
   if (pstr[plen] != EOP ) syserr(UNKNOWN);
}


/* clear program */
clearpgm ()
{
   setpstr(0,EOL);
   setpstr(1,EOP);
   plen = 1;
}


/* set variables (either create new or clear old) */
initvar (clr)
   int clr;
{
   int i;

   if (!clr)
      { memarg(&var_nam, var_len*4);
        memarg(&var_ptr, var_len*4);
        memarg(&var_num, var_len*2); }

   i=0;
   while (i<var_len)
      { if (clr)
           { if (var_nam[i] != 0L) free(var_nam[i]);
             if (var_ptr[i] != 0L) free(var_ptr[i]); }
        var_nam[i] = 0L;
        var_ptr[i] = 0L;
        var_num[i] = 0;
        i++; }
}


/* set loop structure (create new or clear old) */
initlp (clr)
   int clr;
{
   if (clr)
      while (lp_cnt > 0)
         { lp_cnt--;
           free(lp_var[lp_cnt]); lp_var[lp_cnt]=0L; }
   else
      { memarg(&lp_ptr, lp_len*2);
        memarg(&lp_end, lp_len*2);
        memarg(&lp_inc, lp_len*2);
        memarg(&lp_var, lp_len*4); }
}


/* compares 2 strings */
streql (s,t)
   char *s, *t;
{
   while (*s == *t) { if (*s == '\0') return 1; s++; t++; }
   return 0;
}


/* hash function */
hash (var)
    char var[];
{
   int i, h;
   char ch;

   h=0; i=0;
   while ((ch=var[i++])!='\0') h+=ch;

   return (h % var_len);
}


/*
 * QL specific commands begin here
 */


/* initializes the display */
initQLscr ()
{
   window(stdout,WINWIDTH,WINHEIGHT,WINSIDE,WINTOP);
   paper(stdout,7);
   ink(stdout,0);
   border(stdout,1,0);
   mode(4);
}


/* gets input ZX81 style */
ZXgets (lstr, io)
   char lstr[]; int io;
{
   int i, j, k, r, c, in;
   char ch, curs, cl[LINELEN];

   dquote = 1;        /* show double quote: "" */
   endrow = CURSROW;  /* set end row to edit mode */
   inprow = CURSROW;

   k=0; in=0;
   if (io)
   {
      r = row; c = col; curs = 'L';
      cl[k++] = '%'; cl[k++] = curs;
   }
   else
   {
      curs = 'K';
      if (lstr[0] == '\0')
      {
         cl[k++] = '%'; cl[k++] = curs;
      }
      else
      {
         j=0;
         while ((ch=lstr[j])!='\0')
         {
            if (!in) if (ch == ' ') { in = k; cl[k++] = '%'; cl[k++] = curs; }
            
            cl[k++] = ch; j++;
         }
      }
   }
   cl[k++] = '\0';
         
   while (1)
   {
      ZXclear(cl, 0);
      ZXprintf(cl);
      ZXprintf("  ");  /* two trailing spaces to remove prev char w/deleting */

      ch = toupper(getchar());

      if (ch == 200)
      {
         if (in < (k-5) && cl[in+2] == '\\')
            { cl[in] = '\\'; cl[in+1] = cl[in+3]; cl[in+2] = cl[in+4];
              cl[in+3] = '%'; cl[in+4] = curs; in = in + 3; }
         else if (in < (k-4) && cl[in+2] == '%')
            { cl[in+1] = cl[in+3]; cl[in+3] = curs; in = in + 2; }
         else if (in < (k-3))
            { cl[in] = cl[in+2]; cl[in+1] = '%'; cl[in+2] = curs; in = in + 1; }
      }
      else if (ch == 194)
      {
         if (in > 2 && cl[in-3] == '\\')
            { i=in-3; while (i<(k-2)) { cl[i] = cl[i+3]; i++; } in = in - 3; k = k - 3; }
         else if (in > 1 && cl[in-2] == '%')
            { i=in-2; while (i<(k-1)) { cl[i] = cl[i+2]; i++; } in = in - 2; k = k - 2;}
         else if (in > 0)
            { i=in-1; while (i<k) { cl[i] = cl[i+1]; i++; } in = in - 1; k = k - 1; }
      }
      else if (ch == 192)
      {
         if (in > 2 && cl[in-3] == '\\')
            { cl[in+1] = cl[in-1]; cl[in] = cl[in-2]; cl[in-1] = '\\';
              cl[in-3] = '%'; cl[in-2] = curs; in = in - 3; }
         else if (in > 1 && cl[in-2] == '%')
            { cl[in+1] = cl[in-1]; cl[in-1] = curs; in = in - 2; }
         else if (in > 0)
            { cl[in+1] = cl[in-1]; cl[in-1] = '%'; cl[in] = curs; in = in - 1; }
      }
      else if (ch == 19)   /* Ctrl-S quits editing (returns empty) */
      {
         if (io || k != 3) 
            { k=0; goto end_input; }
         continue;
      }
      else if (ch == 10)
      {
         i=0; j=0; k=0;
         while ((ch=cl[j]) != '\0')
         {
            if (j++ == in) { j++; continue; }

            if (!k) if( ch == ' ') continue;        /* skip all leading spaces */

            lstr[i++] = ch;
            if (ch != ' ') k = i;  /* keep track of first trailing space or end */
         }

         end_input:
         lstr[k] = EOL; /* add end with new-line */
         lstr[k+1] = '\0';

         break;
      }
      else 
      {
         if (k > LINEMAX) continue;

         if (ch != '\\') if (ch != '%')
            { i=k; while (i>=(in+1)) cl[i--] = cl[i];
              cl[in++] = ch; k = k + 1;
              continue; }

         cl[in+1] = 'G';
         ZXclear(cl, 0);
         ZXprintf(cl);
         
         if (ch == '\\')
            { i=k+2; while (i>=(in+3)) { cl[i] = cl[i-3]; i--; } cl[in++] = '\\';
              cl[in++] = toupper(getchar()); cl[in++] = toupper(getchar());
              k = k + 3; }
         else
            { i=k+1; while (i>=(in+2)) { cl[i] = cl[i-2]; i--; } cl[in++] = '%'; 
              cl[in++] = toupper(getchar()); k = k + 2; }

         cl[in+1] = curs;
      }
   }

   if (io)
   {
      lstr[k] = '\0';
      ZXclear(cl, 1);
      row = r; col = c;
   }
   else
   {
      /* if we are 'E'diting then no cls */
      i=0; while (lstr[i]==' ') i++;
      if (lstr[i] != 'E') cls(stdout); else ZXclear(cl, 1);

      row = FIRSTROW; col = FIRSTCOL;
   }

   dquote = 0;        /* back to single quote: "      */
   endrow = LASTROW;  /* reset end row to screen mode */
}


/* gets number of ZX characters */
ZXlen(lstr)
   char lstr[];
{
   int i, l;
   char ch;

   i = 0; l = 0;
   while ((ch=lstr[i++])!='\0')
      { switch(ch) { case '\\': if (lstr[i]=='"') { l+=dquote; } i++; case '%': i++; } l++; }

   return l;
}


/* clears edit window */
ZXclear (lstr, all)
   char lstr[]; int all;
{
   int i, j, r;


   r = CURSROW - (((j=ZXlen(lstr)) / COLLEN) * 8);
   if ((r < inprow) || all)
   {
      block(stdout, MAXROW, 2, 2, r-2, 7); /* clear the top 2 lines */

      inprow = r;
      if (!all) j = (1 + ((CURSROW - r) / 8)) * COLLEN;

      i=0; row = r; col = FIRSTCOL - 8;
      while (i++<j)
      {
         ZXadjustcurs();
         ZXputc(' ', 0, 7);
      }
   }
   else
      r = inprow;

   row = r; col = FIRSTCOL;
}


/* prints in ZX81 font */
ZXprintf (lstr)
   char lstr[];
{
   char ch;
   int i;

   i=0;
   while ((ch=lstr[i++])!='\0')
   {
      if (row > endrow) return NOMORE;
      switch (ch)
      {
         case EOL:
            ZXdowncurs();
            continue;
         case '\\':
            ZXgraphics(lstr[i++], lstr[i++]);
            break;
         case '%':
            ZXputc(lstr[i++], 7, 0);
            break;
         default:
            ZXputc(ch, 0, 7);
      }
      ZXadjustcurs();
   }
   return NOERR;
}


/* prints single ZX81 font character */
ZXputc (ch, ink, pap)
   char ch; int ink, pap;
{
   block(stdout, 8, 8, col, row, pap);

   switch (ch)
   {
     /* LETTERS */

      case 'A':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 5, col+1, row+2, ink);
         block(stdout, 1, 5, col+6, row+2, ink);
         block(stdout, 4, 1, col+2, row+4, ink);
         return;
      case 'B':
         block(stdout, 5, 6, col+1, row+1, ink);
         block(stdout, 4, 1, col+2, row+2, pap);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 4, 2, col+2, row+4, pap);
         block(stdout, 1, 2, col+6, row+4, ink);
         return;
/*->*/case 'G':
         block(stdout, 3, 1, col+4, row+4, ink);
      case 'C':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 4, col+1, row+2, ink);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 1, 1, col+6, row+5, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;
      case 'D':
         block(stdout, 4, 1, col+1, row+1, ink);
         block(stdout, 1, 4, col+1, row+2, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 2, col+6, row+3, ink);
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 4, 1, col+1, row+6, ink);
         return;
      case 'E':
         block(stdout, 5, 1, col+2, row+6, ink);
      case 'F':
         block(stdout, 6, 1, col+1, row+1, ink);
         block(stdout, 1, 5, col+1, row+2, ink);
         block(stdout, 4, 1, col+2, row+3, ink);
         return;
/*->  case 'G': <-- moved before letter C */
      case 'H':
         block(stdout, 1, 6, col+1, row+1, ink);
         block(stdout, 4, 1, col+2, row+3, ink);
         block(stdout, 1, 6, col+6, row+1, ink);
         return;
      case 'I':
         block(stdout, 5, 1, col+2, row+1, ink);
         letter_I:
         block(stdout, 1, 4, col+4, row+2, ink);
         block(stdout, 5, 1, col+2, row+6, ink);
         return;
/*->*/case 'U':
         block(stdout, 1, 3, col+1, row+1, ink);
      case 'J':  /* adds extra block call 3 vs 4 */
         block(stdout, 1, 2, col+1, row+4, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         block(stdout, 1, 5, col+6, row+1, ink);
         return;
      case 'K':
         block(stdout, 1, 6, col+1, row+1, ink);
         block(stdout, 1, 1, col+5, row+1, ink);
         block(stdout, 1, 1, col+4, row+2, ink);
         block(stdout, 2, 1, col+2, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 1, 1, col+6, row+6, ink);
         return;
      case 'L':
         block(stdout, 1, 6, col+1, row+1, ink);
/*->*/case '_':
         block(stdout, 5, 1, col+2, row+6, ink);
         return;
      case 'M':
         block(stdout, 1, 1, col+2, row+2, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 2, 1, col+3, row+3, ink);
         goto letter_N;
      case 'N':
         block(stdout, 1, 1, col+2, row+2, ink);
         block(stdout, 1, 1, col+3, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 1, col+5, row+5, ink);
         letter_N:
         block(stdout, 1, 6, col+1, row+1, ink);
         block(stdout, 1, 6, col+6, row+1, ink);
         return;
/*->*/case '0': /* zero */
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 1, col+4, row+3, ink);
         block(stdout, 1, 1, col+3, row+4, ink);
         block(stdout, 1, 1, col+2, row+5, ink);
      case 'O':
         letter_O:
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 4, col+1, row+2, ink);
         block(stdout, 1, 4, col+6, row+2, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;
/*->*/case 'R':
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 1, 1, col+6, row+6, ink);
      case 'P':
         block(stdout, 5, 1, col+1, row+1, ink);
         block(stdout, 1, 5, col+1, row+2, ink);
         block(stdout, 4, 1, col+2, row+4, ink);
         block(stdout, 1, 2, col+6, row+2, ink);
         return;
      case 'Q':
         block(stdout, 1, 1, col+3, row+4, ink);
         block(stdout, 1, 1, col+4, row+5, ink);
         goto letter_O;
/*->  case 'R': <-- moved before letter P */
      case 'S':
         block(stdout, 4, 1, col+2, row+1, ink);
         letter_S1:
         block(stdout, 1, 1, col+1, row+2, ink);
         block(stdout, 1, 1, col+1, row+5, ink);
         letter_S2:
         block(stdout, 4, 1, col+2, row+3, ink);
         block(stdout, 1, 2, col+6, row+4, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;
      case 'T':
         block(stdout, 7, 1, col, row+1, ink);
         block(stdout, 1, 5, col+3, row+2, ink);
         return;
/*->  case 'U': <-- moved before letter J */
      case 'V':
         block(stdout, 1, 4, col+1, row+1, ink);
         block(stdout, 1, 4, col+6, row+1, ink);
         block(stdout, 1, 1, col+2, row+5, ink);
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 2, 1, col+3, row+6, ink);
         return;
      case 'W':
         block(stdout, 1, 5, col+1, row+1, ink);
         block(stdout, 1, 5, col+6, row+1, ink);
         block(stdout, 2, 1, col+3, row+5, ink);
         block(stdout, 1, 1, col+2, row+6, ink);
         block(stdout, 1, 1, col+5, row+6, ink);
         return;
      case 'X':
         block(stdout, 1, 1, col+1, row+1, ink);
         block(stdout, 1, 1, col+6, row+1, ink);
         block(stdout, 1, 1, col+2, row+2, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 2, 2, col+3, row+3, ink);
         block(stdout, 1, 1, col+2, row+5, ink);
         block(stdout, 1, 1, col+5, row+5, ink);
         block(stdout, 1, 1, col+1, row+6, ink);
         block(stdout, 1, 1, col+6, row+6, ink);
         return;
      case 'Y':
         block(stdout, 1, 1, col, row+1, ink);
         block(stdout, 1, 1, col+6, row+1, ink);
         block(stdout, 1, 1, col+1, row+2, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 1, col+2, row+3, ink);
         block(stdout, 1, 1, col+4, row+3, ink);
         block(stdout, 1, 3, col+3, row+4, ink);
         return;
      case 'Z':
         block(stdout, 6, 1, col+1, row+1, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 1, col+4, row+3, ink);
         block(stdout, 1, 1, col+3, row+4, ink);
         block(stdout, 1, 1, col+2, row+5, ink);
         block(stdout, 6, 1, col+1, row+6, ink);
         return;

      /* NUMBERS */

/*->  case '0': <-- moved before letter O */
      case '1':
         block(stdout, 2, 1, col+3, row+1, ink);
         block(stdout, 1, 1, col+2, row+2, ink);
         goto letter_I;
      case '2':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 1, col+1, row+2, ink);
         block(stdout, 1, 2, col+6, row+2, ink);
         block(stdout, 4, 1, col+2, row+4, ink);
         block(stdout, 1, 1, col+1, row+5, ink);
         block(stdout, 6, 1, col+1, row+6, ink);
         return;
      case '3':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 1, col+1, row+2, ink);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 2, 1, col+4, row+3, ink);
         block(stdout, 1, 2, col+6, row+4, ink);
         block(stdout, 1, 1, col+1, row+5, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;
      case '4':
         block(stdout, 1, 6, col+4, row+1, ink);
         block(stdout, 1, 1, col+3, row+2, ink);
         block(stdout, 1, 1, col+2, row+3, ink);
         block(stdout, 1, 1, col+1, row+4, ink);
         block(stdout, 6, 1, col+1, row+5, ink);
         return;
      case '5':
         block(stdout, 6, 1, col+1, row+1, ink);
         goto letter_S1;
      case '6':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 4, col+1, row+2, ink);
         goto letter_S2;
      case '7':
         block(stdout, 6, 1, col+1, row+1, ink);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 1, 1, col+5, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 2, col+3, row+5, ink);
         return ;
      case '8':
         block(stdout, 4, 6, col+2, row+1, ink);
         block(stdout, 6, 4, col+1, row+2, ink);
         block(stdout, 4, 1, col+2, row+2, pap);
         block(stdout, 4, 2, col+2, row+4, pap);
         block(stdout, 1, 1, col+1, row+3, pap);
         block(stdout, 1, 1, col+6, row+3, pap);
         return ;
      case '9':
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 2, col+1, row+2, ink);
         block(stdout, 1, 4, col+6, row+2, ink);
         block(stdout, 4, 1, col+2, row+4, ink);
         block(stdout, 4, 1, col+2, row+6, ink);
         return;

      /* SYMBOLS */

      case '"':
         block(stdout, 1, 2, col+3, row+1, ink);
         block(stdout, 1, 2, col+6, row+1, ink);
         return;
      case '#':
         block(stdout, 3, 1, col+3, row+1, ink);
         block(stdout, 1, 4, col+2, row+2, ink);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 4, 1, col+1, row+3, ink);
         block(stdout, 6, 1, col+1, row+6, ink);
         return;
      case '$':
         block(stdout, 5, 5, col+2, row+2, ink);
         block(stdout, 4, 1, col+3, row+3, pap);
         block(stdout, 4, 1, col+2, row+5, pap);
         block(stdout, 1, 7, col+4, row+1, ink);
         return;
      case ':':
         block(stdout, 1, 1, col+3, row+3, ink);
         block(stdout, 1, 1, col+3, row+6, ink);
          return;
      case '(':
         block(stdout, 1, 1, col+5, row+1, ink);
         block(stdout, 1, 4, col+4, row+2, ink);
         block(stdout, 1, 1, col+5, row+6, ink);
         return;
      case ')':
         block(stdout, 1, 1, col+2, row+1, ink);
         block(stdout, 1, 4, col+3, row+2, ink);
         block(stdout, 1, 1, col+2, row+6, ink);
         return;
      case '>':
         block(stdout, 1, 1, col+2, row+2, ink);
         block(stdout, 1, 1, col+3, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 1, col+3, row+5, ink);
         block(stdout, 1, 1, col+2, row+6, ink);
         return;
      case '<':
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 1, col+4, row+3, ink);
         block(stdout, 1, 1, col+3, row+4, ink);
         block(stdout, 1, 1, col+4, row+5, ink);
         block(stdout, 1, 1, col+5, row+6, ink);
         return;
      case '=':
         block(stdout, 5, 1, col+2, row+3, ink);
         block(stdout, 5, 1, col+2, row+5, ink);
         return;
      case '*':
         block(stdout, 1, 3, col+4, row+3, ink);
         block(stdout, 1, 1, col+3, row+2, ink);
         block(stdout, 1, 1, col+5, row+2, ink);
         block(stdout, 1, 1, col+3, row+6, ink);
         block(stdout, 1, 1, col+5, row+6, ink);
      case '-':
         sym_minus:
         block(stdout, 5, 1, col+2, row+4, ink);
         return;
      case '+':
         block(stdout, 1, 5, col+4, row+2, ink);
         goto sym_minus;
/*->  case '_': <-- moved to letter L */
      case '/':
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 1, 1, col+5, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 1, col+3, row+5, ink);
         block(stdout, 1, 1, col+2, row+6, ink);
         return;
      case ';':
         block(stdout, 1, 1, col+3, row+2, ink);
         block(stdout, 1, 2, col+3, row+5, ink);
         block(stdout, 1, 1, col+2, row+7, ink);
         return;
      case ',':
         block(stdout, 1, 2, col+4, row+5, ink);
         block(stdout, 1, 1, col+3, row+7, ink);
         return;
      case '.':
         block(stdout, 2, 2, col+3, row+5, ink);
         return;
      case ' ':
         return;

      /* UNKOWN is ? */

      default:
         block(stdout, 4, 1, col+2, row+1, ink);
         block(stdout, 1, 1, col+1, row+2, ink);
         block(stdout, 1, 1, col+6, row+2, ink);
         block(stdout, 1, 1, col+5, row+3, ink);
         block(stdout, 1, 1, col+4, row+4, ink);
         block(stdout, 1, 1, col+4, row+6, ink);
         return;
   }
}


/* prints ZX81 graphics character */
ZXgraphics (ch1, ch2)
   char ch1, ch2;
{
   switch (ch1)
   {
      case '"':
         if (ch2 == '"')
            { ZXputc ('"', 0, 7);
              if (dquote)
                 { ZXadjustcurs();
                   if (row <= endrow)
                      ZXputc ('"', 0, 7); }
              return; }
         break;
      case '#':
         if (ch2 == '#')
            { ZXgrey(1, 1, 0, 7); return; }
         break;
      case '~':
         if (ch2 == '~')
            { ZXgrey(1, 0, 0, 7); return; }
         break;
      case ',':
         if (ch2 == ',')
            { ZXgrey(0, 1, 0, 7); return; }
         break;
      case '@':
         if (ch2 == '@')
            { ZXgrey(1, 1, 7, 0); return; }
         break;
      case '!':
         if (ch2 == '!')
            { ZXgrey(1, 0, 7, 0); return; }
         break;
      case ';':
         if (ch2 == ';')
            { ZXgrey(0, 1, 7, 0); return; }
         break;
      case ' ':
         if (ch2 == ':')
            { ZXblock(0, 1, 0, 1); return; }
         if (ch2 == '\'')
            { ZXblock(0, 1, 0, 0); return; }
         if (ch2 == '.')
            { ZXblock(0, 0, 0, 1); return; }
         if (ch2 == ' ')
            { ZXblock(0, 0, 0, 0); return; }
               return;
         break;
      case ':':
         if (ch2 == ':')
            { ZXblock(1, 1, 1, 1); return; }
         if (ch2 == '\'')
            { ZXblock(1, 1, 1, 0); return; }
         if (ch2 == '.')
            { ZXblock(1, 0, 1, 1); return; }
         if (ch2 == ' ')
            { ZXblock(1, 0, 1, 0); return; }
         break;
      case '\'':
         if (ch2 == ':')
            { ZXblock(1, 1, 0, 1); return; }
         if (ch2 == '\'')
            { ZXblock(1, 1, 0, 0); return; }
         if (ch2 == '.')
            { ZXblock(1, 0, 0, 1); return; }
         if (ch2 == ' ')
            { ZXblock(1, 0, 0, 0); return; }
         break;
      case '.':
         if (ch2 == ':')
            { ZXblock(0, 1, 1, 1); return; }
         if (ch2 == '\'')
            { ZXblock(0, 1, 1, 0); return; }
         if (ch2 == '.')
            { ZXblock(0, 0, 1, 1); return; }
         if (ch2 == ' ')
            { ZXblock(0, 0, 1, 0); return; }
         break;
   }

   /* Question mark for unknown characters */
   ZXputc('?', 0, 7);
}


/* create specfic ZX81 grey character */
ZXgrey (up, lo, ink, pap)
   int up, lo, ink, pap;
{
   int pat;

   if (pap == 0)
      pat = 246;  /* even stipple pattern */
   else
      pat = 248;  /* odd stipple pattern */

   block(stdout, 8, 8, col, row, pap);

   if (up)
   {
      block(stdout, 8, 4, col, row, pat);
   }
   if (lo)
   {
      block(stdout, 8, 4, col, row+4, pat);
   }
}


/* create ZX81 block characters */
ZXblock (ul, ur, ll, lr)
   int ul, ur, ll, lr;
{
   block(stdout, 8, 8, col, row, 7);

   if (ul)
      block(stdout, 4, 4, col, row, 0);
   if (ur)
      block(stdout, 4, 4, col+4, row, 0);
   if (ll)
      block(stdout, 4, 4, col, row+4, 0);
   if (lr)
      block(stdout, 4, 4, col+4, row+4, 0);
}


/* move cursor down */
ZXdowncurs ()
{
   col = FIRSTCOL;
   row = row + 8;
}


/* adjust cursor for end of line/screen */
ZXadjustcurs ()
{
   col = col + 8;
   if (col > LASTCOL)
   {
      col = FIRSTCOL;
      row = row + 8;
   }
}


/****************************** expr_c ******************************/


/* gets a variable from command line */
getvar (var)
   char var[];
{
   int  i, op;
   char ch;

   while (str[sp] == ' ') sp++;
   if (!isalpha(str[sp])) return NOVAL;
   i=0;
   while (i<VARLEN)
      if (isalpha(ch=str[sp]) || isdigit(ch) || (ch==' '))
         { if (ch != ' ')
              { var[i++] = ch; sp++; }
           else 
              { while (str[sp]==' ') sp++;
                switch (str[sp])
                   { case 'T':
                       if (getkey("THEN") == NOERR) { sp = sp - 4; goto end_var; }
                       if (getkey("TO") == NOERR) { sp = sp - 2; goto end_var; }
                       break;
                    case 'S':
                       if (getkey("STEP") == NOERR) { sp = sp - 4; goto end_var; }
                       break;
                    case 'A':
                       if (getkey("AND") == NOERR) { sp = sp - 3; goto end_var; }
                       break;
                    case 'O':
                       if (getkey("OR") == NOERR) { sp = sp - 2; goto end_var; }
         }    }    }
      else
         { if (ch == '$' || ch == '(')
              { var[i++] = ch; sp++; if (i == VARLEN) break; }
           end_var: var[i] = '\0'; return NOERR; }

   return BADARG;
}


/* finds var in hashtable */
findvar(var,vp)
   char var[]; int *vp;
{
   int i, j, *num;
   addr *nam, *ptr;

   i = hash(var);
   
   j=i;
   do {
      if (var_nam[j] == 0L)
         { *vp = j; return NOVAL; }
      else if (streql(var_nam[j],var))
         { *vp = j; return NOERR; }
      j=(j+1)%var_len;
   } while (j!=i);


   /* if hash is filled up we double space */
   nam=var_nam; ptr=var_ptr; num=var_num;

   var_len+=VARSIZE;
   initvar(CREATE);

   i=0;
   while (i<(var_len-VARSIZE))
      { findvar(nam[i],&j);
        var_nam[j]=nam[i]; var_ptr[j]=ptr[i]; var_num[j]=num[i]; i++; }
   free(nam); free(ptr); free(num);

   return findvar(var,vp);
}


/* sets a variable */
storevar (var,ind,val_i,val_s)
   char var[]; int ind, val_i; char val_s[];
{
   addr mem;
   char ch;
   int  i, err;

   if ((err=findvar(var,&i)) == NOVAL)
      { memarg(&mem,strlen(var)+1);
        var_nam[i]=mem; strcpy(mem,var); }

   if ((ch=var[strlen(var)-1]) == '$')
      { memarg(&mem,strlen(val_s)+1);
        if (err == NOERR) free(var_ptr[i]);
        var_ptr[i] = mem; strcpy(mem,val_s); }
   else if (ch == '(')
      setintp(var_ptr[i],ind,val_i);
   else
      var_num[i] = val_i;

   return NOERR;
}


/* retrievs value of a variable */
fetchvar (var,ind,val_i,val_s)
   char var[]; int ind, *val_i; addr *val_s;
{
   char ch;
   int i, j, k, dim, len, err;

   if (findvar(var,&i) != NOERR) return BADARG;

   if ((ch=var[strlen(var)-1]) == '$')
      { doSTR = 1;
        memarg(val_s,strlen(var_ptr[i])+1);
        strcpy(*val_s,var_ptr[i]); *val_i = atoi(*val_s); }
   else 
      { if (ch == '(')
           *val_i = getintp(var_ptr[i],ind);
        else
           *val_i = var_num[i];
        memarg(val_s,NUMLEN); itoa(*val_i,*val_s); }

   return NOERR;
}


/* retrievs array index for DIM'ed array */
getindex (var,val_i,val_s)
   char var[]; int *val_i; addr *val_s;
{
   int i, j, ind, len, dim;

   if (findvar(var,&i) != NOERR) return BADARG;

   j = 1; ind = 0; len = 0; dim = getintp(var_ptr[i],0);
   while (1) {
      if (getexpr(val_i,val_s) == NOERR) free(*val_s); else return BADARG;
      len = getintp(var_ptr[i],j++);
      if (*val_i < 1 || *val_i > len) return BADARG;
      if (j > dim) break;
      ind = (ind + *val_i - 1) * len;
      if (getsym(",") != NOERR) return BADARG;
   }
   if (getsym(")") != NOERR) return BADARG;

   *val_i = dim + ind + *val_i;
   memarg(val_s,NUMLEN); itoa(*val_i,*val_s);

   return NOERR;
}


/* gets number from command line */
getnum (val_i,val_s)
   int *val_i; addr *val_s;
{
   char ch;
   int i;

   while (str[sp] == ' ') sp++;
   if (!isdigit(str[sp])) return NOVAL;
   i=0; memarg(val_s,NUMLEN);
   while (isdigit(ch=str[sp]))
      { setchp(*val_s,i++,ch); sp++; }
   setchp(*val_s,i,'\0');
   *val_i = atoi(*val_s);

   return NOERR;
}


/* gets quoted string from command line */
getstr (val_i,val_s)
   int *val_i; addr *val_s;
{
   char ch;
   int i, tp, len;

   while (str[sp] == ' ') sp++;
   if (str[sp] == '"') { doSTR = 1; tp = ++sp; } else return NOVAL;
   while((ch=str[tp])!='"') if (ch=='\\') tp=tp+3; else { tp++; if (ch==EOL) return BADCHAR; }
   len = tp-sp; memarg(val_s,len+1);
   i=0; while (i<len) setchp(*val_s,i++,str[sp++]);
   setchp(*val_s,i,'\0'); sp++;
   *val_i = atoi(*val_s);

   return NOERR;   
}


/* parses substring expression */
getsubstr (tmp_i,tmp_s,val_i,val_s)
   int *tmp_i; addr *tmp_s;
   int *val_i; addr *val_s;
{
   char ch;
   int err, from, to, i, j, cp, in, idx, len;

   if ((err=getrange(&from,&to,(len=strlen(*tmp_s)))) == NOVAL)
      { *val_i=*tmp_i; *val_s=*tmp_s; return NOERR; }
   else if (err != NOERR)
      { free(*tmp_s); return BADARG; }
   
   if (to < from)
      { j=0; memarg(val_s,1); }
   else
      {
        cp=0;
        while (cp<2)  /* first pass see how much space, no copy */
        {
           if (cp) memarg(val_s,j);

           i=0; j=0; in=0; idx=0;
           while (i<len)
           {
              switch (ch=getchp(*tmp_s,i))
              {
                 case '%':
                    idx++;
                    if (in || from == idx) 
                       { if (cp) 
                            { setchp(*val_s,j++, ch); i++;
                              setchp(*val_s,j++,getchp(*tmp_s,i)); i++; }
                         else
                            { i=i+2; j=j+2; }
                         if (idx == to) goto end_idx;
                         in = 1; }
                    else
                       i=i+2;
                    break;
                 case '\\':
                    idx++;
                    if (in || from == idx)
                       { if (cp) 
                            { setchp(*val_s,j++, ch); i++;
                              setchp(*val_s,j++,getchp(*tmp_s,i)); i++;
                              setchp(*val_s,j++,getchp(*tmp_s,i)); i++; }
                         else
                            { i=i+3; j=j+3; }
                         if (idx == to) goto end_idx;
                         in = 1; }
                    else
                       i=i+3;
                    break;
                 default:
                    idx++;
                    if (in || from == idx)
                       { if (cp) setchp(*val_s,j,ch);
                         j++;
                         if (idx == to) goto end_idx;
                         in = 1; }
                    i++;
              }
           }
           end_idx:
           cp++;
        }
      }

   setchp(*val_s,j,'\0');  *val_i = atoi(*val_s);

   free(*tmp_s);
   return NOERR;
}


/* gets range of string */
getrange (from,to,len)
   int *from, *to, len;
{
   char *tmp_s;
   int  noTO;

   if (getsym("(") != NOERR) return NOVAL;

   if (noTO=(getkey("TO") != NOERR))
      { if (getexpr(from,&tmp_s) ==NOERR) free(tmp_s); else return BADARG; }
   else
      *from = 1;

   if (!noTO || (getkey("TO") == NOERR))
      { if (getexpr(to,&tmp_s) == NOERR) free(tmp_s); else *to = len; }
   else
      *to = *from;

   if (getsym(")") != NOERR) return BADARG;

   if ((*from < 1) || (*to < 1)) return BADARG;
   if ((*from > len) || (*to > len)) { if (*from <= *to) return BADARG; }
   return NOERR;
}


/* gets value from command line */
getval (val_i,val_s)
   int *val_i; addr *val_s;
{
   addr tmp_s;
   int  tmp_i, err;
   char var[VARLEN];

   if (getsym("(") == NOERR)  /* eval subexpression */
      { if (getexpr(val_i,val_s) != NOERR) return BADARG;
        return getsym(")"); }
   else if ((err=evalfun(val_i,val_s)) != NOVAL)
      return err;
   else if ((err=getnum(val_i,val_s)) != NOVAL)
      return err;
   else if ((err=getstr(&tmp_i,&tmp_s)) != NOVAL)
      { if (err != NOERR) return err;
        return getsubstr(&tmp_i,&tmp_s,val_i,val_s); }
   else if (getvar(var) == NOERR)
      { if (var[strlen(var)-1] == '(')
           { if (getindex(var,&tmp_i,&tmp_s) == NOERR) free(tmp_s); else return BADARG;
             return fetchvar(var,tmp_i,val_i,val_s); }
        if (fetchvar(var,0,&tmp_i,&tmp_s) != NOERR) return BADARG;
           return getsubstr(&tmp_i,&tmp_s,val_i,val_s); }

   return BADARG;
}


/* gets {chr$(),len()} functions and unary {-,not} expressions */
evalfun (val_i,val_s)
   int *val_i; addr *val_s;
{
   int i, inv;

   delim = ws;  /* function delimter different from other keywords */
   if (getsym("-") == NOERR) {
      if (getval(val_i,val_s) != NOERR) return BADARG;
      *val_i = -*val_i;
   }
   else switch (str[sp])
   {
      case 'A':
         if (getkey("ABS") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            *val_i = abs(*val_i);
            break;
         }
         goto noval;
      case 'C':
         if (getkey("CHR$") == NOERR) {
            if (getval(val_i,val_s) != NOERR) { return BADARG; } free(*val_s);
            if (*val_i >> 8) return BADARG;

            i=0; inv = *val_i > 127; memarg(val_s,2+inv);
            if (inv) { *val_i = *val_i - 128; setchp(*val_s,i++,'%'); }

            if (*val_i == 0)
               *val_i = 32;
            else if (*val_i <11 || *val_i > 63)
               { *val_i = 15; i=0; }
            else if (*val_i >= 38 )
               *val_i = *val_i + 27;  /* 'A' is 65 for ASCII - 38 for ZX81 */
            else if (*val_i >= 28)
               *val_i = *val_i + 20;  /* '0' is 48 for ASCII - 28 for ZX81 */
            else
               *val_i = sym[*val_i-11];

            doSTR = 1; setchp(*val_s,i++,*val_i); setchp(*val_s,i,'\0');
            goto skip_free;  /* we want to keep string value */
         }
         if (getkey("CODE") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            doSTR = 0; *val_i = getchp(*val_s,0);
      
            if (*val_i == '\0')
               *val_i = 0;
            else if (*val_i == '\\')
               { if (getchp(*val_s,1)==':' && getchp(*val_s,2)==':') *val_i=128; else *val_i=67; }
            else
               { if (*val_i != '%') inv = 0; else { inv = 128; *val_i = getchp(*val_s,1); }

                 if (*val_i == 32)
                    *val_i = 0 + inv;
                 else if (*val_i >= 48 && *val_i <= 57)
                    *val_i = *val_i - 20 + inv;   /* '0' is 48 for ASCII - 28 for ZX81 */
                 else if (*val_i >= 65 && *val_i <= 90)
                    *val_i = *val_i - 27 + inv;   /* 'A' is 65 for ASCII - 38 for ZX81 */
                 else if (*val_i < 192)  /* 192 & up are QL control characters we want */
                    { i=0; while (i<17) if (sym[i] == *val_i) break; else i++;
                      if (i == 17) *val_i = 67; else *val_i = i + 11 + inv; } }
            break;
         }
         goto noval;
      case 'I':
         if (getkey("INKEY$") == NOERR) {
            i = 0; memarg(val_s,2); trap3(key, ret);
            if (!ret[0]) { setchp(*val_s,i++,toupper(getchar())); } setchp(*val_s,i,'\0');
            if (getchp(*val_s,0)==32) { setchp(*val_s,0,252); }  /* for space 32 map to 252 */
            doSTR = 1; *val_i = atoi(*val_s);
            goto skip_free;  /* want to keep string value */
         }
         if (getkey("INT") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            break;
         }
         goto noval;
      case 'L':
         if (getkey("LEN") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            doSTR = 0; *val_i = ZXlen(*val_s);
            break;
         }
         goto noval;
      case 'N':
         if (getkey("NOT") == NOERR) {
            if (getexpr(val_i,val_s) != NOERR) return BADARG;
            *val_i = !*val_i;
            break;
         }
         goto noval;
      case 'P':
         if (getkey("PI") == NOERR) {
            *val_i = PI;
            break;
         }
         goto noval;
      case 'R':
         if (getkey("RND") == NOERR) {
            seed = ((seed * P1) + P2) % N;
            *val_i = seed; *val_i = abs(*val_i);   /* int from long then its absolute value */
            if (getsym("*") == NOERR)
               { if (getval(&i,val_s) != NOERR) return BADARG;
                 *val_i = *val_i % i; if (i<0) *val_i = -*val_i; }   /* hack until float */
            break;
         }
         goto noval;
      case 'S':
         if (getkey("SGN") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            *val_i = sign(*val_i);
            break;
         }
         if (getkey("STR$") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            doSTR = 1;
            goto skip_free;  /* we want to keep string value */
         }
         goto noval;
      case 'V':
         if (getkey("VAL") == NOERR) {
            if (getval(val_i,val_s) != NOERR) return BADARG;
            doSTR = 0;
            break;
         }
      default:
         noval:
   	 delim = spc; return NOVAL;
   }

   /* free string since function only returns numeric value */
   free(*val_s); memarg(val_s,NUMLEN); itoa(*val_i,*val_s);
   skip_free:

   delim = spc;  /* reset delimeter to space */
   return NOERR;
}


/* parses ** expressions */
evalpow (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i, pow, i;
   addr right_s;

   if (getval(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '*' : {
            if (str[sp+1] != '*') return NOERR;
            sp = sp + 2;
            if (getval(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            i = 0; pow = 1;
            while (i++ < right_i)
               pow *= *left_i;

            *left_i = pow;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free (*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}


/* parses {*,/} expressions */
evalmul (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evalpow(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '*' : {
            sp++;
            if (evalpow(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i * right_i;
            break;
         }
         case '/' : {
            sp++;
            if (evalpow(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i / right_i;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}


/* parses {+,-} expressions */
evaladd (left_i,left_s)
   int *left_i; addr *left_s;
{
   int i, j, right_i;
   addr right_s, tmp_s;

   doSTR = 0;  /* casting is true for each add */
   if (evalmul(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      j=0; 
      while (str[sp] == ' ') sp++;
      switch (str[sp])
      {
         case '+' : {
            sp++;
            if (evalmul(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }

            /* concatenate string if we grabbed one */
            if (doSTR)
               { memarg(&tmp_s,strlen(*left_s)+strlen(right_s)+1);
                 i=0; while (getchp(*left_s,i)!='\0') setchp(tmp_s,j++,getchp(*left_s,i++));
                 i=0; while (getchp(right_s,i)!='\0') setchp(tmp_s,j++,getchp(right_s,i++));
                 setchp(tmp_s,j,'\0'); free(*left_s); *left_s = tmp_s;
                 *left_i = atoi(*left_s); }
            else
               *left_i = *left_i + right_i;
            break;
         }
         case '-' : {
            sp++;
            if (evalmul(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
            *left_i = *left_i - right_i;
            break;
         }
         default  : {
            return NOERR;
         }
      }
      free(right_s);
      if (j==0)
         { free(*left_s); memarg(left_s,NUMLEN); itoa(*left_i,*left_s); }
   }
}


/* parses {=,>,<,<>,<=,>=} expressions */
evaleql (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evaladd(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      if (getsym("=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) == 0);
         else
            *left_i = *left_i == right_i;
      }
      else if (getsym("<=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) <= 0);
         else
            *left_i = *left_i <= right_i;
      }
      else if (getsym("<>") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) != 0);
         else
            *left_i = *left_i != right_i;
      }
      else if (getsym("<") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) < 0);
         else
            *left_i = *left_i < right_i;
      }
      else if (getsym(">=") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) >= 0);
         else
            *left_i = *left_i >= right_i;
      }
      else if (getsym(">") == NOERR) {
         if (evaladd(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         if (doSTR)
            *left_i = (strcmp(*left_s,right_s) > 0);
         else
            *left_i = *left_i > right_i;
      }
      else {
         doSTR = 0; /* reset string casting to false */
         return NOERR;
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}


/* parses {and,or} expressions */
evalbool (left_i,left_s)
   int *left_i; addr *left_s;
{
   int right_i;
   addr right_s;

   if (evaleql(left_i,left_s) != NOERR) return BADARG;
   while (1)
   {
      if (getkey("AND") == NOERR) {
         if (evaleql(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         *left_i = *left_i && right_i;
      }
      else if (getkey("OR") == NOERR) {
         if (evaleql(&right_i,&right_s) != NOERR) { free(*left_s); return BADARG; }
         *left_i = *left_i || right_i;
      }
      else {
         return NOERR;
      }
      free(*left_s); free(right_s);
      memarg(left_s,NUMLEN); itoa(*left_i,*left_s);
   }
}

